import { join } from 'path';

import createRunnerServer from './helpers/createRunnerServer';
import generateKey from './helpers/generateKey';
import IRunner from './types/IRunner';
import ISystem from './types/ISystem';
import Messenger from './types/Messenger';

const Runner = async (system: ISystem, messageHandler: Messenger, installDirectory: string): Promise<IRunner> => {
  const services: Array<{ name: string; key: string; port: number; stop: () => Promise<void>; running: boolean }> = [];
  const serverPort = 8000;
  let nextPort = serverPort + 1;

  const getDirectory = (name: string) => join(installDirectory, `colosseum-${name}`);

  const getService = (name: string) => services.find(service => service.name === name);

  const start = async (name: string, key: string, port: number) => {
    const directory = join(getDirectory(name), 'node_modules', name);
    return system.startProcess('node', [
      join(__dirname, './indexService'),
      `--path=${directory}`,
      `--key=${key}`,
      `--port=${port}`,
      `--host=http://localhost:${serverPort}`,
    ]);
  };

  const register = async (name: string) => {
    const key = await generateKey();
    const port = nextPort++;
    const stop = await start(name, key, port);
    services.push({ name, key, port, stop, running: false });
  };

  const registerInstalled = async () => {
    const installed = await system.listDirectories(installDirectory);
    for (const name of installed) {
      await register(name.replace(/^colosseum-/, ''));
    }
  };

  const add = async (name: string, url: string) => {
    if (getService(name)) {
      throw new Error(`Service with name ${name} already exists.`);
    }

    await system.createProject(installDirectory, `colosseum-${name}`);
    await system.install(getDirectory(name), url);
    await register(name);
  };

  const remove = async (name: string) => {
    await system.deleteDirectory(getDirectory(name));
    await services.splice(services.findIndex(x => x.name === name), 1);
  };

  const running = async (key: string) => {
    const service = services.find(s => s.key === key);

    if (!service) {
      throw new Error('Invalid key');
    }

    service.running = true;
  };

  const getServices = async () =>
    services.map(x => ({
      name: x.name,
      running: x.running,
    }));

  const update = async (name: string) => {
    const service = getService(name);
    if (!service) {
      throw new Error(`Service ${name} not registered.`);
    }

    await service.stop();
    await system.update(getDirectory(name), name);
    const stop = await start(name, service.key, service.port);
    service.stop = stop;
  };

  const message: Messenger = async (name, messageName, payload) => {
    const service = getService(name);
    if (!service) {
      return { success: false, response: `Service ${name} not registered.` };
    }

    const url = `http://localhost:${service.port}/${messageName}/${service.key}`;
    const { body, status } = await system.post(url, payload);

    return { success: status === 200, response: body };
  };

  const load = async () => {
    await registerInstalled();

    const authenticate = (key: string) => Boolean(services.find(service => service.key === key));
    const server = createRunnerServer(authenticate, messageHandler);

    await system.serve(server, serverPort);
  };

  await load();

  return {
    add,
    getServices,
    message,
    remove,
    running,
    update,
  };
};

export default Runner;
