import * as bodyParser from 'body-parser';
import * as cookieParser from 'cookie-parser';
import * as express from 'express';
import { Request, Response } from 'express';
import { readFileSync } from 'fs-extra';
import { join } from 'path';

import { generateToken, verifyToken } from './helpers/authentication';
import Messenger from './types/Messenger';

const Server = async (message: Messenger, secret: string, adminPasswordHash: string) => {
  const server = express();

  server.use(bodyParser.json());
  server.use(cookieParser());

  server.get('/status', (req, res) => {
    res.send('👌');
  });

  server.get('/authenticate', (req, res) => {
    res.send(readFileSync(join(__dirname, '../static/authenticate.html'), 'utf-8'));
  });

  server.post('/authenticate', async (req, res) => {
    const password = req.body.password;
    const token = await generateToken(secret, adminPasswordHash, password);
    if (token) {
      res.json({ token });
    } else {
      res.status(401).send('🖕');
    }
  });

  const handleMessage = async (req: Request, res: Response) => {
    try {
      const token = req.params.token || req.cookies.token;
      if (await verifyToken(secret, token)) {
        const response = await message(req.params.service, req.params.message, req.body);
        res.status(response.success ? 200 : 500);
        res.send(response.response);
      } else {
        res.status(401).send('🖕');
      }
    } catch (ex) {
      res.status(500).send(ex.stack);
    }
  };

  server.post('/:service/:message/:token', handleMessage);

  server.post('/:service/:message', handleMessage);

  server.get('/:service*', async (req, res) => {
    const token = req.cookies.token;
    if (await verifyToken(secret, token)) {
      const path = req.path.substr(req.params.service.length + 1);
      const payload = {
        path,
      };
      const response = await message(req.params.service, 'get', payload);

      if (response.success) {
        res.status(response.response.status || 200);
        Object.keys(response.response.headers || {}).forEach(header => {
          res.setHeader(header, response.response.headers[header]);
        });
        res.send(response.response.body);
      } else {
        res.status(500).send(response.response);
      }
    } else {
      res.status(401).send('🖕');
    }
  });

  return server;
};

export default Server;
