import { Express } from 'express';
import { emptyDirSync, ensureDirSync, lstatSync, removeSync } from 'fs-extra';
import { join } from 'path';
import * as request from 'request';

import ISystem from '../types/ISystem';
import * as NPM from './npm';
import { start } from './process';

const ls = require('ls'); // tslint:disable-line no-var-requires
const ncu = require('npm-check-updates'); // tslint:disable-line no-var-requires

const createProject = async (directory: string, name: string) => {
  const projectDirectory = join(directory, name);
  emptyDirSync(projectDirectory);

  await NPM.init(projectDirectory);
};

const deleteDirectory = async (directory: string) => {
  removeSync(directory);
};

const ensureDirectory = async (directory: string) => {
  ensureDirSync(directory);
};

const install = async (directory: string, url: string) => {
  await NPM.installDependency(directory, url);
};

const listDirectories = async (directory: string) => {
  return ls(join(directory, '*'))
    .filter((file: any) => lstatSync(file.full).isDirectory())
    .map((file: any) => file.name);
};

const post = (url: string, requestBody: any) =>
  new Promise<{ body: string; status: number }>((resolve, reject) => {
    request.post(url, { json: requestBody }, (error, response, responeBody) => {
      if (error) {
        reject(error);
      } else {
        resolve({ status: response.statusCode, body: responeBody });
      }
    });
  });

const serve = (server: Express, port: number) =>
  new Promise<void>(resolve =>
    server.listen(port, () => {
      resolve();
    }),
  );

const startProcess = async (command: string, args: string[]) => {
  const childProcess = await start(command, args, process.cwd(), console.log, console.error, () => null);
  return childProcess.stop;
};

const update = async (directory: string, name: string) => {
  await deleteDirectory(join(directory, 'node_modules'));
  await NPM.install(directory);
  await ncu.run({
    packageFile: join(directory, 'package.json'),
    silent: true,
  });
};

const System: ISystem = {
  createProject,
  deleteDirectory,
  ensureDirectory,
  install,
  listDirectories,
  post,
  serve,
  startProcess,
  update,
};

export default System;
